/* This file is part of the Imote2 Codeloader.
 *
 * The Imote2 Codeloader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Imote2 Codeloader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Imote2 Codeloader. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * Image.java
 *
 * @author Markus Engel
 * Copyright (c) 2009,2010 Networked Systems Group, University of Kaiserslautern
 */
package imote;

import java.io.EOFException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * Represents some information of an image, given by its header
 * @author m_engel
 */
public class Image {
	/** first magic number of the header (old and new images) */
	final private static long HEADER_MAGIC = 0xcafebabel;
	/** final magic number at EOF (old and new images) */
	final private static long END_MAGIC = 0xdeadc0del;

	/**
	 * represents the header information of an image
	 */
	private static class NewHeader {
		/** offset of header inside an image */
		final private static long HEADER_OFFSET = 0x40l;
		/** size of a header inside an image */
		final private static int HEADER_SIZE = 8 * 4;
		/** 1st header field: first magic */
		private long magic;
		/** 2nd header field: image header version, encoded as
		 * (major,minor) = (version >> 16, version & 0xffff) */
		private long version;
		/** 3rd header field: image size */
		private long size;
		/** 4th header field: Virtual Memory Address. Image expects to be
		 * located there (either physically or being remapped by the mmu) */
		private long vma;
		/** 5th header field: entry point of the image, absolute address
		 * with respect to the vma */
		private long entry;
		/** 6th header field: description pointer of the image, absolute
		 * address with respect to the vma */
		private long description;

		/**
		 * Read an images header
		 * @param input file which represents the image, is not rewound to the
		 *  previous file pointer
		 * @throws IOException whenever something bad happens while reading
		 * @throws Exception if the image is too short and cannot be read
		 */
		public NewHeader(RandomAccessFile input) throws IOException, Exception {
			/* check image length */
			if (input.length() < HEADER_OFFSET + HEADER_OFFSET) {
				throw new Exception("Image too short, cannot read header");
			}

			/* setup a byte array where the header is read in */
			byte[] hdr = new byte[HEADER_SIZE];
			/* seek to the header offset inside the file */
			input.seek(HEADER_OFFSET);
			/* read the header */
			input.read(hdr);
			/* wrap the byte array into a ByteBuffer to read the fields */
			ByteBuffer b = ByteBuffer.wrap(hdr);
			/* byteorder is little endian */
			b.order(ByteOrder.LITTLE_ENDIAN);

			this.magic = intToULong(b.getInt());
			this.version = intToULong(b.getInt());
			this.size = intToULong(b.getInt());
			this.vma = intToULong(b.getInt());
			this.entry = intToULong(b.getInt());
			this.description = intToULong(b.getInt());
		}
	}

	/**
	 * represents the header information of an old image
	 */
	private static class OldHeader {
		/** offset of header inside an image */
		final private static long HEADER_OFFSET = 0x60l;
		/** size of a header inside an image */
		final private static int HEADER_SIZE = 2 * 4;
		/** 1st header field: first magic */
		private long magic;
		/** 2nd header field: image size */
		private long size;

		/**
		 * Read an old images header
		 * @param input file which represents the image, is not rewound to the
		 *  previous file pointer
		 * @throws IOException whenever something bad happens while reading
		 * @throws Exception if the image is too short and cannot be read
		 */
		public OldHeader(RandomAccessFile input) throws IOException, Exception {
			/* check image length */
			if (input.length() < HEADER_OFFSET + HEADER_OFFSET) {
				throw new Exception("Image too short, cannot read header");
			}

			/* setup a byte array where the header is read in */
			byte[] hdr = new byte[HEADER_SIZE];
			/* seek to the header offset inside the file */
			input.seek(HEADER_OFFSET);
			/* read the header */
			input.read(hdr);
			/* wrap the byte array into a ByteBuffer to read the fields */
			ByteBuffer b = ByteBuffer.wrap(hdr);
			/* byteorder is little endian */
			b.order(ByteOrder.LITTLE_ENDIAN);

			this.magic = intToULong(b.getInt());
			this.size = intToULong(b.getInt());
		}
	}

	/**
	 * Represents the version flag
	 */
	public enum VersionFlag {
		/** old image with short header */
		VERSION_OLD,
		/** new image with full header */
		VERSION_NEW
	}

	public static class Version {
		/** Version flag */
		private VersionFlag versionFlag;
		/** major version number (if applicable) */
		private int major;
		/** minor version number (if applicable) */
		private int minor;

		public int getMajor() {
			return this.major;
		}

		public int getMinor() {
			return this.minor;
		}

		private void setVersion(long version) {
			this.major = (int) (version >> 16);
			this.minor = (int) (version & 0xffff);
		}
	}
	/** vma of image */
	private long vma = 0;
	/** relocated entry of image */
	private long entry = 0;
	/** size of image */
	private long size = 0;
	/** version of image */
	private Version version = new Version();
	/** description of image */
	private String description = "";
	/** is the image considered bootable */
	private boolean bootable = false;

	/**
	 * Read image information from a file which represents the image
	 * @param input file which represents the image, is rewound to the current
	 *  file pointer after reading (even in case of an exception)
	 * @throws IOException whenever reading failes
	 * @throws Exception if the image is too short (maybe corrupted)
	 */
	public Image(RandomAccessFile input) throws IOException, Exception {
		/* save file pointer to rewind there */
		long fp = input.getFilePointer();

		try {
			if (this.tryReadNewImage(input)) {
				this.bootable = true;
				this.version.versionFlag = VersionFlag.VERSION_NEW;
			} else if (this.tryReadOldImage(input)) {
				this.bootable = true;
				this.version.versionFlag = VersionFlag.VERSION_OLD;
			}
		} finally {
			/* in every case: rewind the file */
			input.seek(fp);
		}
	}

	private boolean tryReadNewImage(RandomAccessFile input) throws IOException,
		Exception {
		NewHeader hdr;

		try {
			/* read header */
			hdr = new NewHeader(input);
		} catch (Exception e) {
			return false;
		}

		if (hdr.magic != HEADER_MAGIC) {
			/* magic1 check did not pass => not a new image */
			return false;
		}

		/* check if we can read the end magic */
		if (hdr.size != input.length()) {
			/* we cannot...image too short (maybe corrupted) */
			//throw new Exception("Cannot read image end magic, image corrupted");
			return false;
		}

		/* seek end of image and read last word */
		input.seek(hdr.size - 4);
		/* read end magic */
		byte[] buf = new byte[4];
		input.read(buf);
		ByteBuffer b = ByteBuffer.wrap(buf);
		b.order(ByteOrder.LITTLE_ENDIAN);
		long endmagic = intToULong(b.getInt());

		/* check magic */
		if (endmagic != END_MAGIC) {
			/* image is corrupted */
			//throw new Exception("Cannot verify image end magic, image corrupted");
			return false;
		}

		/* read header fields into image info */
		this.size = hdr.size;
		this.vma = hdr.vma;
		this.entry = relocate(hdr.entry);
		this.version.setVersion(hdr.version);
		long desc = this.relocate(hdr.description);
		/* read description */
		this.description = readDescription(input, desc);
		
		return true;
	}

	private boolean tryReadOldImage(RandomAccessFile input) throws IOException,
		Exception {
		OldHeader hdr;

		try {
			/* read header */
			hdr = new OldHeader(input);
		} catch (Exception e) {
			return false;
		}

		if (hdr.magic != HEADER_MAGIC) {
			/* magic1 check did not pass => not an old image */
			return false;
		}

		/* check if we can read the end magic */
		if (hdr.size + 4 != input.length()) {
			/* we cannot...image too short (maybe corrupted) */
			//throw new Exception("Cannot read image end magic, image corrupted");
			return false;
		}

		/* seek end of image and read last word */
		input.seek(hdr.size);
		/* read end magic */
		byte[] buf = new byte[4];
		input.read(buf);
		ByteBuffer b = ByteBuffer.wrap(buf);
		b.order(ByteOrder.LITTLE_ENDIAN);
		long endmagic = intToULong(b.getInt());

		/* check magic */
		if (endmagic != END_MAGIC) {
			/* image is corrupted */
			//throw new Exception("Cannot verify image end magic, image corrupted");
			return false;
		}

		/* read header fields into image info */
		this.size = hdr.size + 4;
		this.entry = 0x20;
		
		return true;
	}

	private static long intToULong(int i) {
		long l = i;

		if (l < 0) {
			l += (1l << 32);
		}

		return l;
	}

	/**
	 * relocate an address with respect to the vma
	 * @param addr address
	 * @return relocated address
	 */
	private long relocate(long addr) {
		return addr - this.vma;
	}

	/**
	 * read description of an image
	 * @param input file which represents the image. it is not rewound.
	 * @param addr address of the description within the file
	 * @return description as a string
	 * @throws IOException whenever reading fails
	 * @throws Exception if image is too short (end of string cannot be found)
	 */
	private static String readDescription(RandomAccessFile input, long addr)
		throws IOException, Exception {
		StringBuilder str = new StringBuilder(30);

		/** seek to the given address */
		input.seek(addr);

		while (true) {
			/* try to read one character */
			char c;
			try {
				c = (char) input.readByte();
			} catch (EOFException e) {
				throw new Exception("Image too short, cannot read description");
			}
			if (c == 0) {
				/* terminating byte found, all done */
				break;
			}
			/* append character */
			str.append(c);
		}

		/* return the read string */
		return str.toString();
	}

	/** wether image is bootable by bootloader
	 *
	 * @return true iff image is bootable
	 */
	public boolean isBootable() {
		return this.bootable;
	}

	/** wether image has an description in its header
	 *
	 * @return true iff image has an description
	 */
	public boolean hasDescription() {
		return this.version.versionFlag == VersionFlag.VERSION_NEW &&
			!this.description.isEmpty();
	}

	/** get description of image
	 *
	 * @return description of image
	 * @throws Exception if there is no such description
	 */
	public String getDescription() throws Exception {
		if (!this.hasDescription()) {
			throw new Exception("There is not such description");
		}
		return this.description;
	}

	/** wether image has a version in its header
	 *
	 * @return true iff image has a version
	 */
	public boolean hasVersion() {
		return this.version.versionFlag == VersionFlag.VERSION_NEW;
	}

	/** image version
	 *
	 * @return image version
	 * @throws Exception if there is no such version
	 */
	public Version getVersion() throws Exception {
		if (!this.hasVersion()) {
			throw new Exception("There is not such version");
		}
		return this.version;
	}
}
